from pika import BaseConnection, BlockingConnection, ConnectionParameters
from typing import Any, Dict, Iterator, List, Tuple, Union
from kh_common.utilities import getFullyQualifiedClassName
from kh_common.config.credentials import message_queue
from kh_common.config.repo import name, short_hash
from kh_common.logging import getLogger, Logger
from pika.channel import Channel
from traceback import format_tb
from types import TracebackType
from sys import exc_info


class Receiver :

	def __init__(self) :
		self._route: str = message_queue['routing_key']
		self._connection_info: Dict[str, Union[str, int]] = message_queue['connection_info']
		self._channel_info: Dict[str, Union[float, bool]] = message_queue['channel_info']
		self._exchange_info: Dict[str, str] = message_queue.get('exchange_info')
		self.logger: Logger = getLogger()


	def consumer(self) -> Iterator[Any] :
		yield from self._recv()


	def receiveAll(self) -> List[Any] :
		return list(self._recv())


	def receiveJson(self, forcelist:bool=False) -> Union[List[Dict[str, Any]], Iterator[Dict[str, Any]]] :
		if forcelist :
			return list(map(json.loads, self._recv()))
		else :
			return map(json.loads, self._recv())


	def _recv(self) -> Iterator[Any] :
		connection: Union[BaseConnection, None] = None
		try :
			# returns a list of all messages retrieved from the message queue
			connection = BlockingConnection(ConnectionParameters(**self._connection_info))
			channel: Channel = connection.channel()

			if self._exchange_info :
				channel.exchange_declare(**self._exchange_info)
				name: str = channel.queue_declare(self._route).method.queue
				channel.queue_bind(routing_key=self._route, queue=name, exchange=self._exchange_info['exchange'])

			else :
				channel.queue_declare(self._route)
				name: str = self._route

			it: Iterator[Tuple[Any]] = channel.consume(name, **self._channel_info)

			ack: int = -1
			for method_frame, _, body in it :
				if body :
					yield body
					ack = max(ack, method_frame.delivery_tag)
				else :
					break

			if ack >= 0 :
				channel.basic_ack(delivery_tag=ack, multiple=True)

			channel.cancel()
		
		except :
			self.logger.warning('message queue crashed unexpectedly.', exc_info=True)

		finally :
			# don't channel.cancel here since, if it fails, we want the messages to remain in the queue
			try :
				if connection :
					connection.close()

			except Exception as e :
				self.logger.warning('unexpected exception occurred during message queue close.', exc_info=e)



from starlette.middleware.trustedhost import TrustedHostMiddleware
from kh_common.server.middleware.auth import KhAuthMiddleware
from kh_common.server.middleware.cors import KhCorsMiddleware
from kh_common.exceptions.base_error import BaseError
from kh_common.utilities.json import json_stream
from fastapi.responses import PlainTextResponse, UJSONResponse
from starlette.exceptions import ExceptionMiddleware
from kh_common.config.constants import environment
from kh_common.exceptions import jsonErrorHandler
from fastapi.responses import Response
from fastapi import FastAPI, Request
from typing import Iterable
from ujson import dumps


NoContentResponse = Response(None, status_code=204)


def ServerApp(
	auth: bool = True,
	auth_required: bool = True,
	cors: bool = True,
	max_age: int = 86400,
	allowed_hosts: Iterable[str] = [
		'localhost',
		'127.0.0.1',
		'*.kheina.com',
		'kheina.com',
	],
	allowed_origins: Iterable[str] = [
		'localhost',
		'127.0.0.1',
		'dev.kheina.com',
		'kheina.com',
	],
	allowed_methods: Iterable[str] = [
		'GET',
		'POST',
	],
	allowed_headers: Iterable[str] = [
		'accept',
		'accept-language',
		'authorization',
		'cache-control',
		'content-encoding',
		'content-language',
		'content-length',
		'content-security-policy',
		'content-type',
		'cookie',
		'host',
		'location',
		'referer',
		'referrer-policy',
		'set-cookie',
		'user-agent',
		'www-authenticate',
		'x-frame-options',
		'x-xss-protection',
	],
	exposed_headers: Iterable[str] = [
		'authorization',
		'cache-control',
		'content-type',
		'cookie',
		'set-cookie',
		'www-authenticate',
	],
) -> FastAPI :
	app = FastAPI()
	app.add_middleware(ExceptionMiddleware, handlers={ Exception: jsonErrorHandler }, debug=False)
	app.add_exception_handler(BaseError, jsonErrorHandler)

	allowed_protocols = ['http', 'https'] if environment.is_local() else ['https']

	if cors :
		app.add_middleware(
			KhCorsMiddleware,
			allowed_origins = set(allowed_origins),
			allowed_protocols = set(allowed_protocols),
			allowed_headers = allowed_headers,
			allowed_methods = allowed_methods,
			exposed_headers = exposed_headers,
			max_age = max_age,
		)

	if allowed_hosts :
		app.add_middleware(TrustedHostMiddleware, allowed_hosts=set(allowed_hosts))

	if auth :
		app.add_middleware(KhAuthMiddleware, required=auth_required)

	return app


class JsonResponse(Response) :
	
	def __init__(self, serializable_body, *args, **kwargs) :
		super().__init__(dumps(json_stream(serializable_body)), *args, **kwargs)
		self.headers['content-type'] = 'application/json'



from typing import Any, Callable, Dict, Iterable, List, Tuple, Union
from kh_common.models import KhUser
from pydantic import BaseModel
from datetime import datetime
from decimal import Decimal
from enum import Enum
from uuid import UUID


_conversions: Dict[type, Callable] = {
	datetime: str,
	Decimal: float,
	tuple: lambda x : list(map(json_stream, filter(None, x))),
	filter: lambda x : list(map(json_stream, filter(None, x))),
	set: lambda x : list(map(json_stream, filter(None, x))),
	list: lambda x : list(map(json_stream, filter(None, x))),
	dict: lambda x : dict(zip(map(str, x.keys()), map(json_stream, x.values()))),
	zip: lambda x : dict(zip(map(str, x.keys()), map(json_stream, x.values()))),
	Enum: lambda x : x.name,
	UUID: lambda x : x.hex,
	KhUser: lambda x : {
		'user_id': x.user_id,
		'scope': json_stream(x.scope),
		'token': {
			'expires': json_stream(x.token.expires),
			'guid': json_stream(x.token.guid),
			'data': x.token.data,
		} if x.token else None,
	},
	BaseModel: lambda x : json_stream(x.dict()),
}


def json_stream(item: Any) -> Any :
	if isinstance(item, str) :
		return item
	for cls in type(item).__mro__ :
		if cls in _conversions :
			return _conversions[cls](item)
	return item



from os import environ, path, listdir
from ujson import load as json_load
from typing import Any, Dict


environment: str = environ.get('ENVIRONMENT', 'LOCAL').lower()

env_vars: Dict[str, Any] = { }

# dynamically load local credentials
if path.isdir('credentials') :
	for filename in listdir('credentials') :
		if filename.endswith('.json') :
			config: Dict[str, Dict[str, Any]] = json_load(open(f'credentials/{filename}'))
			c: Dict[str, Any] = config.get(environment) or config.get('prod')
			if not c :
				continue

		# add other file type logic here

		env_vars.update(c)
		del filename, config, c

# add the variables from the environment to the module
locals().update(env_vars)

# delete extraneous data
del env_vars, environment, environ, path, listdir, json_load



import time

def pretty_time(time) :
	test_time = abs(time)
	if test_time > 5400:  # 90 minutes in seconds
		return str(round(time / 3600, 2)) + 'hr'
	elif test_time > 90:
		return str(round(time / 60, 2)) + 'min'
	elif test_time < 0.000001:
		return str(round(time * 1000000000, 2)) + 'ns'
	elif test_time < 0.001:
		return str(round(time * 1000000, 2)) + 'μs'
	elif test_time < 1:
		return str(round(time * 1000, 2)) + 'ms'
	return str(round(time, 2)) + 's'

def test(*args, iterations=1000000, base=None, **kwargs) :
	start = 0
	end = 0
	functions = { a.__name__: a for a in args }
	functions.update(kwargs)

	bad = False
	match = True
	outputs = []
	for label, function in functions.items() :
		try :
			outputs.append(function())
			print(f'[{label}][output] {outputs[-1]}')
			if len(outputs) >= 2 and outputs[-1] != outputs[-2] :
				match = False
		except Exception as e :
			print(f'[{label}][error] threw {e.__class__.__name__}: {e}')
			bad = True

	if bad : return
	elif not match : print('[warning] not all outputs match. (test may not be accurate)')

	del bad
	del match
	del outputs

	if base:
		start = time.process_time()
		for i in range(iterations) :
			base()
		end = time.process_time()
	else:
		start = time.process_time()
		for i in range(iterations) :
			try : pass
			except : pass
		end = time.process_time()
	base = end - start
	print(f'[base] {pretty_time(base)} for {iterations:,} iterations ({pretty_time(base / iterations)} per iteration)')

	for label, function in functions.items() :
		start = time.process_time()
		for i in range(iterations) :
			try : function()
			except : pass
		end = time.process_time()
		function_total = (end - start) - base
		print(f'[{label}] {pretty_time(function_total)} for {iterations:,} iterations ({pretty_time(function_total / iterations)} per iteration)')



from typing import Any, Callable, Dict, Iterator, List, Tuple, Union
from kh_common.utilities import flatten, getFullyQualifiedClassName
from kh_common.config.repo import short_hash, name as repo_name
from google.cloud import logging as google_logging
from kh_common.utilities.json import json_stream
from google.auth import compute_engine
from traceback import format_tb
from types import ModuleType
import logging


class TerminalAgent :

	loggable: Tuple[type] = (str, int, float, None)

	def __init__(self) -> None :
		import time
		import json
		self.time: ModuleType = time
		self.json: ModuleType = json

	def log_text(self, log: str, severity:str='INFO') -> None :
		print('[' + self.time.asctime(self.time.localtime(self.time.time())) + ']', severity, '>', log)

	def log_struct(self, log: Dict[str, Any], severity:str='INFO') -> None :
		for i in flatten(log) :
			if not isinstance(i, TerminalAgent.loggable) :
				print('WARNING:', i, 'may not be able to be logged.')
		print('[' + self.time.asctime(self.time.localtime(self.time.time())) + ']', severity, '>', self.json.dumps(log, indent=4))


class LogHandler(logging.Handler) :

	logging_available = True

	def __init__(self, name: str, *args: Tuple[Any], structs:Tuple[type]=(dict, list, tuple), **kwargs:[str, Any]) -> None :
		logging.Handler.__init__(self, *args, **kwargs)
		self._structs: Tuple[type] = structs
		try :
			if not LogHandler.logging_available :
				raise ValueError('logging unavailable.')
			credentials: compute_engine.credentials.Credentials = compute_engine.Credentials()
			logging_client: google_logging.client.Client = google_logging.Client(credentials=credentials)
			self.agent: google_logging.logger.Logger = logging_client.logger(name)
		except :
			LogHandler.logging_available = False
			self.agent: TerminalAgent = TerminalAgent()


	def emit(self, record: logging.LogRecord) -> None :
		if record.args and isinstance(record.msg, str) :
			record.msg: str = record.msg % record.args
		if record.exc_info :
			e: Exception = record.exc_info[1]
			refid = getattr(e, 'refid', None)
			errorinfo: Dict[str, Any] = {
				'error': f'{getFullyQualifiedClassName(e)}: {e}',
				'stacktrace': list(map(str.strip, format_tb(record.exc_info[2]))),
				'refid': refid.hex if refid else None,
				**json_stream(getattr(e, 'logdata', { })),
			}
			if isinstance(record.msg, dict) :
				errorinfo.update(json_stream(record.msg))
			else :
				errorinfo['message'] = record.msg
			self.agent.log_struct(errorinfo, severity=record.levelname)
		else :
			if isinstance(record.msg, self._structs) :
				self.agent.log_struct(record.msg, severity=record.levelname)
			else :
				self.agent.log_text(str(record.msg), severity=record.levelname)


Logger: type = logging.Logger


def getLogger(name: Union[str, None]=None, level:int=logging.INFO, filter:Callable=lambda x : x, disable:List[str]=[], **kwargs:[str, Any]) -> Logger :
	name: str = name or f'{repo_name}.{short_hash}'
	for loggerName in disable :
		logging.getLogger(loggerName).propagate = False
	logging.root.setLevel(logging.NOTSET)
	handler: LogHandler = LogHandler(name, level=level)
	handler.addFilter(filter)
	logging.root.handlers.clear()
	logging.root.addHandler(handler)
	return logging.getLogger(name)



from kh_common.utilities import getFullyQualifiedClassName


class Hashable :

	_hash = 0

	def __init__(self) :
		self._hash = hash(f"<class '{getFullyQualifiedClassName(self)}' {Hashable._hash}>")
		Hashable._hash += 1


	def __hash__(self) :
		return self._hash



from aiohttp import ClientTimeout, request as async_request


async def post(url: str, json=None, data=None, timeout=30, retries=3) :
	for attempt in range(retries) :
		try :
			async with async_request('POST', url, json=json, data=data, timeout=ClientTimeout(timeout)) as response :
				response.raise_for_status()
				return response

		except :
			if attempt + 1 >= retries :
				raise


async def get(url: str, timeout=30, retries=3) :
	for attempt in range(retries) :
		try :
			async with async_request('GET', url, timeout=ClientTimeout(timeout)) as response :
				response.raise_for_status()
				return response

		except :
			if attempt + 1 >= retries :
				raise



from enum import Enum
from time import time


class TimeUnit(Enum) :
	planck = 5.39e-44
	yoctosecond = 1e-24
	jiffy = 3e-24
	zeptosecond = 1e-21
	attosecond = 1e-18
	femtosecond = 1e-15
	svedberg = 1e-13
	picosecond = 1e-12
	nanosecond = 1e-9
	shake = 1e-8
	microsecond = 1e-6
	millisecond = 1e-3
	second = 1
	decasecond = 10
	minute = 60
	moment = 90
	hectosecond = 100
	decaminute = 600
	ke = 864
	kilosecond = 1000
	hour = 3600
	hectominute = 6000
	kilominute = 60000
	day = 86400
	week = 604800
	megasecond = 1000000
	fortnight = 1209600
	month = 2592000
	quarter = 7776000
	season = 7776000
	quadrimester = 10368000
	semester = 1555200
	year = 31536000
	common_year = 31536000
	tropical_year = 31556925.216
	gregorian = 31556952
	sidereal_year = 31558149.7635456
	leap_year = 31622400
	biennium = 63072000
	triennium = 94608000
	quadrennium = 126144000
	olympiad = 126144000
	lustrum = 157680000
	decade = 315360000
	indiction = 473040000
	gigasecond = 1000000000
	jubilee = 1576800000
	century = 3153600000
	millennium = 31536000000
	terasecond = 1000000000000
	megannum = 31536000000000
	petasecond = 1000000000000000
	galactic_year = 7253279999999999
	aeon = 31536000000000000
	exasecond = 1000000000000000000
	zettasecond = 1000000000000000000000
	yottasecond = 1000000000000000000000000


class Timer :

	def __init__(self) :
		self._start = None
		self._end = None


	def start(self) :
		self._start = time()
		return self


	def end(self) :
		self._end = time()
		return self


	def elapsed(self, unit: TimeUnit = TimeUnit.second) :
		end = self._end or time()
		return (end - self._start) / unit.value



from datetime import timezone, datetime as pydatetime
from typing import Union


class datetime(pydatetime) :

	def fromtimestamp(timestamp: Union[int, float], timezone: timezone = timezone.utc) :
		return pydatetime.fromtimestamp(timestamp, timezone)


	def now(timezone: timezone = timezone.utc) :
		return pydatetime.now(timezone)


from base64 import urlsafe_b64encode, urlsafe_b64decode
from typing import Union


def b64encode(bytestring: bytes) -> bytes :
	return urlsafe_b64encode(bytestring).strip(b'=')


def b64decode(bytestring: Union[str, bytes]) -> bytes :
	return urlsafe_b64decode(bytestring + (4 - len(bytestring) % 4) * (b'=' if isinstance(bytestring, bytes) else '='))



from starlette.types import ASGIApp, Receive, Send, Scope as request_scope
from kh_common.auth import AuthToken, KhUser, retrieveAuthToken, Scope
from kh_common.exceptions.http_error import HttpError, Unauthorized
from kh_common.exceptions import jsonErrorHandler
from starlette.requests import Request


class KhAuthMiddleware:

	def __init__(self, app: ASGIApp, required: bool = True) -> None :
		self.app = app
		self.auth_required = required


	async def __call__(self, scope: request_scope, receive: Receive, send: Send) -> None :
		if scope['type'] not in { 'http', 'websocket' } :
			raise NotImplementedError()

		request: Request = Request(scope, receive, send)

		try :
			token_data: AuthToken = retrieveAuthToken(request)

			scope['user'] = KhUser(
				user_id=token_data.user_id,
				token=token_data,
				scope={ Scope.user } | set(map(Scope.__getitem__, token_data.data.get('scope', []))),
			)

		except HttpError as e :
			if isinstance(e, Unauthorized) and self.auth_required :
				response = jsonErrorHandler(request, e)
				await response(scope, receive, send)
				return

			scope['user'] = KhUser(
				user_id=None,
				token=None,
				scope={ Scope.default },
			)

		await self.app(scope, receive, send)



from psycopg2.extensions import connection as Connection, cursor as Cursor
from typing import Any, Callable, Dict, List, Tuple, Union
from psycopg2 import Binary, connect as dbConnect
from psycopg2.errors import ConnectionException
from kh_common.logging import getLogger, Logger
from kh_common.config.credentials import db
from kh_common.timing import Timer


class SqlInterface :

	def __init__(self, long_query_metric: float = 1, conversions: Dict[type, Callable] = { }) -> None :
		self.logger: Logger = getLogger()
		self._sql_connect()
		self._long_query = long_query_metric
		self._conversions: Dict[type, Callable] = {
			tuple: list,
			bytes: Binary,
			**conversions,
		}


	def _sql_connect(self) -> None :
		try :
			self._conn: Connection = dbConnect(**db)

		except Exception as e :
			self.logger.critical(f'failed to connect to database!', exc_info=e)

		else :
			self.logger.info('connected to database.')


	def _convert_item(self, item: Any) -> Any :
		item_type = type(item)
		if item_type in self._conversions :
			return self._conversions[item_type](item)
		return item


	def query(self, sql: str, params:Tuple[Any]=(), commit:bool=False, fetch_one:bool=False, fetch_all:bool=False, maxretry:int=2) -> Union[None, List[Any]] :
		if self._conn.closed :
			self._sql_connect()

		params = tuple(map(self._convert_item, params))
		try :
			cur: Cursor = self._conn.cursor()
			
			timer = Timer().start()

			cur.execute(sql, params)

			if commit :
				self._conn.commit()
			else :
				self._conn.rollback()

			if timer.elapsed() > self._long_query :
				self.logger.warning(f'query took longer than {self._long_query} seconds:\n{sql}')

			if fetch_one :
				return cur.fetchone()

			elif fetch_all :
				return cur.fetchall()

		except ConnectionException as e :
			if maxretry > 1 :
				self.logger.warning('connection to db was severed, attempting to reconnect.', exc_info=e)
				self._sql_connect()
				return self.query(sql, params, commit, fetch_one, fetch_all, maxretry - 1)

			else :
				self.logger.critical('failed to reconnect to db.', exc_info=e)
				raise

		except Exception as e :
			self.logger.warning({
				'message': 'unexpected error encountered during sql query.',
				'query': sql,
			}, exc_info=e)
			# now attempt to recover by rolling back
			self._conn.rollback()
			raise

		finally :
			cur.close()


	def transaction(self) :
		return Transaction(self)


	def close(self) -> int :
		self._conn.close()
		return self._conn.closed


class Transaction :

	def __init__(self, sql: SqlInterface) :
		self._sql: SqlInterface = sql
		self.cur: Union[Cursor, None] = None


	def __enter__(self) :
		for _ in range(2) :
			try :
				self.cur: Cursor = self._sql._conn.cursor()
				return self

			except ConnectionException as e :
				self._sql.logger.warning('connection to db was severed, attempting to reconnect.', exc_info=e)
				self._sql._sql_connect()

		raise ConnectionException('failed to reconnect to db.')


	def __exit__(self, exc_type, exc_obj, exc_tb) :
		if exc_type :
			self.rollback()
		self.cur.close()


	def commit(self) :
		self._sql._conn.commit()


	def rollback(self) :
		self._sql._conn.rollback()


	def query(self, sql: str, params:Tuple[Any]=(), fetch_one:bool=False, fetch_all:bool=False) -> Union[None, List[Any]] :
		params = tuple(map(self._sql._convert_item, params))
		try :
			self.cur.execute(sql, params)

			if fetch_one :
				return self.cur.fetchone()

			elif fetch_all :
				return self.cur.fetchall()

		except Exception as e :
			self._sql.logger.warning({
				'message': 'unexpected error encountered during sql query.',
				'query': sql,
			}, exc_info=e)
			raise



from typing import Any, Callable, Dict, Iterable
from inspect import getfullargspec
from math import ceil


def getFullyQualifiedClassName(obj: object) -> str :
	module = getattr(obj, '__module__', None)
	if module and module != str.__module__ :
		return f'{module}.{obj.__class__.__name__}'
	return obj.__class__.__name__


def stringSlice(string: str, start:str=None, end:str=None) -> str :
	if not string : return None
	assert start or end, 'start or end is required'
	start = string.rfind(start) + len(start) if start else None
	end = string.find(end) if end else None
	return string[start:end]


def flatten(it: Iterable[Any]) -> Iterable[Any] :
	if isinstance(it, (tuple, list, set)) :
		for i in it :
			yield from flatten(i)
	elif isinstance(it, dict) :
		for v in it.values() :
			yield from flatten(v)
	else :
		yield it


def int_to_bytes(integer: int) -> bytes :
	return integer.to_bytes(ceil(integer.bit_length() / 8), 'big')


def int_from_bytes(bytestring: bytes) -> int :
	return int.from_bytes(bytestring, 'big')



from kh_common.exceptions.http_error import NotFound
from kh_common.caching import SimpleCache
from kh_common.cwd import setCwd
from typing import List, Tuple
import ujson as json
import os


cwd = setCwd()


@SimpleCache(900)  # 15 minute cache
def secureFolders() -> List[str] :
	try :
		with open('securefolders.json') as folders :
			return json.load(folders)
	except :
		return ['credentials']


def safeJoin(*args: Tuple[str]) -> str :
	path = os.path.realpath(os.path.join(*args))
	if path.startswith(cwd) and all(folder not in path for folder in secureFolders()) and os.path.exists(path) :
		return path
	raise NotFound('The requested resource is not available.')



from typing import Any, Callable, Dict, Hashable, Iterable, Tuple, Union
from inspect import FullArgSpec, getfullargspec, iscoroutinefunction
from collections import defaultdict
from functools import wraps
from math import sqrt
from time import time
from copy import copy


class CalcDict(dict) :

	def __init__(self, default: Callable[[Hashable], Any]) -> None :
		self.default: Callable = default


	def setdefault(self, default: Callable[[Hashable], Any]) -> None :
		self.default = default


	def __missing__(self, key: Hashable) -> Any:
		self[key] = self.default(key)
		return self[key]


_conversions: Dict[type, Callable] = {
	dict: lambda x : tuple((key, x[key]) for key in sorted(x.keys())),
	list: tuple,
}


def _convert_item(item: Any) -> Any :
	if isinstance(item, str) :
		return item
	if isinstance(item, Iterable) :
		return _cache_stream(item)
	for cls in type(item).__mro__ :
		if cls in _conversions :
			return _conversions[cls](item)
	return item


def _cache_stream(stream: Iterable) :
	if isinstance(stream, dict) :
		return tuple((key, _convert_item(stream[key])) for key in sorted(stream.keys()))

	else :
		return tuple(map(_convert_item, stream))


def SimpleCache(TTL_seconds:float=0, TTL_minutes:float=0, TTL_hours:float=0, TTL_days:float=0, ) -> Callable :
	"""
	stores single result for all arguments used to call.
	any arguments/keywords can be used.
	"""
	TTL: float = TTL_seconds + TTL_minutes * 60 + TTL_hours * 3600 + TTL_days * 86400
	del TTL_seconds, TTL_minutes, TTL_hours, TTL_days

	def decorator(func: Callable) -> Callable :
		if iscoroutinefunction(func) :
			@wraps(func)
			async def wrapper(*args: Tuple[Any], **kwargs:Dict[str, Any]) -> Any :
				now: float = time()
				if now > decorator.expire :
					decorator.expire = now + TTL
					decorator.data = await func(*args, **kwargs)
				return copy(decorator.data)

		else :
			@wraps(func)
			def wrapper(*args: Tuple[Any], **kwargs:Dict[str, Any]) -> Any :
				now: float = time()
				if now > decorator.expire :
					decorator.expire = now + TTL
					decorator.data = func(*args, **kwargs)
				return copy(decorator.data)

		return wrapper
	decorator.expire: float = 0
	decorator.data: Any = None
	return decorator


def ArgsCache(TTL_seconds:float=0, TTL_minutes:float=0, TTL_hours:float=0, TTL_days:float=0) -> Callable :
	"""
	stores results for every argument used to call.
	requires all arguments to be hashable, keywords are not included in the cache key.
	"""
	TTL: float = TTL_seconds + TTL_minutes * 60 + TTL_hours * 3600 + TTL_days * 86400
	del TTL_seconds, TTL_minutes, TTL_hours, TTL_days

	def decorator(func: Callable) -> Callable :

		if iscoroutinefunction(func) :
			@wraps(func)
			async def wrapper(*args: Tuple[Any], **kwargs:Dict[str, Any]) -> Any :
				now: float = time()
				key = args

				if decorator.cache :
					i: int = 0
					for expires, itemkey in decorator.keys :
						if expires >= now : break
						del decorator.cache[itemkey]
						i += 1

					if i == len(decorator.keys) :
						decorator.keys.clear()
					elif i :
						decorator.keys = decorator.keys[-len(decorator.cache):]

					if key in decorator.cache :
						return copy(decorator.cache[key])

				data: Any = await func(*args, **kwargs)
				decorator.cache[key]: Any = data
				decorator.keys.append((now + TTL, key))

				return copy(data)

		else :
			@wraps(func)
			def wrapper(*args: Tuple[Any], **kwargs:Dict[str, Any]) -> Any :
				now: float = time()
				key = args

				if decorator.cache :
					i: int = 0
					for expires, itemkey in decorator.keys :
						if expires >= now : break
						del decorator.cache[itemkey]
						i += 1

					if i == len(decorator.keys) :
						decorator.keys.clear()
					elif i :
						decorator.keys = decorator.keys[-len(decorator.cache):]

					if key in decorator.cache :
						return copy(decorator.cache[key])

				data: Any = func(*args, **kwargs)
				decorator.cache[key]: Any = data
				decorator.keys.append((now + TTL, key))

				return copy(data)

		return wrapper

	decorator.cache: Dict[Tuple[Hashable], Any] = { }
	decorator.keys: List[Tuple[Union[float, Hashable]]] = [ ]
	return decorator


def KwargsCache(TTL_seconds:float=0, TTL_minutes:float=0, TTL_hours:float=0, TTL_days:float=0) -> Callable :
	"""
	stores results for every argument used to call.
	recursively converts all arguments/keywords into hashable types, if possible.
	"""
	TTL: float = TTL_seconds + TTL_minutes * 60 + TTL_hours * 3600 + TTL_days * 86400
	del TTL_seconds, TTL_minutes, TTL_hours, TTL_days


	def decorator(func: Callable) -> Callable :

		if iscoroutinefunction(func) :
			@wraps(func)
			async def wrapper(*args: Tuple[Hashable], **kwargs:Dict[str, Hashable]) -> Any :
				cache_key: Tuple[Any] = _cache_stream([args, kwargs])
				now: float = time()

				if decorator.cache :
					i: int = 0
					for expires, key in decorator.keys :
						if expires >= now : break
						del decorator.cache[key]
						i += 1

					if i == len(decorator.keys) :
						decorator.keys.clear()
					elif i :
						decorator.keys = decorator.keys[-len(decorator.cache):]

					if cache_key in decorator.cache :
						return copy(decorator.cache[cache_key])

				data: Any = await func(*args, **kwargs)
				decorator.cache[cache_key]: Any = data
				decorator.keys.append((now + TTL, cache_key))

				return copy(data)

		else :
			@wraps(func)
			def wrapper(*args: Tuple[Hashable], **kwargs:Dict[str, Hashable]) -> Any :
				cache_key: Tuple[Any] = _cache_stream([args, kwargs])
				now: float = time()

				if decorator.cache :
					i: int = 0
					for expires, key in decorator.keys :
						if expires >= now : break
						del decorator.cache[key]
						i += 1

					if i == len(decorator.keys) :
						decorator.keys.clear()
					elif i :
						decorator.keys = decorator.keys[-len(decorator.cache):]

					if cache_key in decorator.cache :
						return copy(decorator.cache[cache_key])

				data: Any = func(*args, **kwargs)
				decorator.cache[cache_key]: Any = data
				decorator.keys.append((now + TTL, cache_key))

				return copy(data)

		return wrapper

	decorator.cache: Dict[Tuple[Any], Any] = { }
	decorator.keys: List[Tuple[Any]] = [ ]
	return decorator


class SumAggregator :
	def __init__(self) :
		self.data = defaultdict(lambda : 0)
	
	def update(self, data: Dict) :
		for k, v in data.items() :
			self.data[k] += v
	
	def result(self) :
		value = self.data.copy()
		self.data.clear()
		return value


class AverageAggregator :
	def __init__(self) :
		self.data = defaultdict(lambda : 0)
		self.count = 0
	
	def update(self, data: Dict) :
		self.count += 1
		for k, v in data.items() :
			self.data[k] += (v - self.data[k]) / self.count
	
	def result(self) :
		value = self.data.copy()
		self.data.clear()
		self.count = 0
		return value


class StandardDeviation :
	def __init__(self, count, avg, q_value) :
		self.average: float = avg
		self.count: int = count
		self.variance: float = q_value / (count - 1)
		self.deviation: float = sqrt(self.variance)


class StandardDeviationAggregator :
	def __init__(self) :
		self.avg = defaultdict(lambda : 0)
		self.variance = defaultdict(lambda : 0)
		self.count = 0

	def update(self, data: Dict) :
		self.count += 1
		for k, v in data.items() :
			prev_avg = self.avg[k]
			self.avg[k] += (v - prev_avg) / self.count
			self.variance[k] += (v - prev_avg) * (v - self.avg[k])

	def result(self) :
		value = {
			k: StandardDeviation(self.count, avg, self.variance[k])
			for k, avg in self.avg.items()
		}
		self.count = 0
		self.avg.clear()
		self.variance.clear()
		return value


class Aggregator :
	Sum: type = SumAggregator
	Average: type = AverageAggregator
	StandardDeviation: type = StandardDeviationAggregator


def Aggregate(TTL_seconds:float=0, TTL_minutes:float=0, TTL_hours:float=0, TTL_days:float=0, exclusions:Iterable[str]=['self'], aggregator:Aggregator=Aggregator.Average) -> Callable :
	"""
	aggregates numeric inputs for a given time span
	"""

	exclusions: Set[str] = set(exclusions)
	aggregator = aggregator()
	TTL: float = TTL_seconds + TTL_minutes * 60 + TTL_hours * 3600 + TTL_days * 86400
	del TTL_seconds, TTL_minutes, TTL_hours, TTL_days

	def decorator(func: Callable) -> Callable :

		arg_spec: FullArgSpec = getfullargspec(func)

		if iscoroutinefunction(func) :
			@wraps(func)
			async def wrapper(*args: Tuple[Any], **kwargs:Dict[str, Any]) -> Any :
				kwargs.update(zip(arg_spec.args, args))
				keys: Set[str] = kwargs.keys() - exclusions
				aggregator.update({
					key: kwargs[key]
					for key in keys
				})

				now: float = time()
				if now > decorator.expire :
					decorator.expire = now + TTL
					kwargs.update(aggregator.result())
					return await func(**kwargs)

		else :
			@wraps(func)
			def wrapper(*args: Tuple[Any], **kwargs:Dict[str, Any]) -> Any :
				kwargs.update(zip(arg_spec.args, args))
				keys: Set[str] = kwargs.keys() - exclusions
				aggregator.update({
					key: kwargs[key]
					for key in keys
				})

				now: float = time()
				if now > decorator.expire :
					decorator.expire = now + TTL
					kwargs.update(aggregator.result())
					return func(**kwargs)

		return wrapper

	decorator.expire: float = time() + TTL
	return decorator



from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
from cryptography.hazmat.primitives import serialization
from tests.utilities.requests import MockResponse
from kh_common.config.constants import auth_host
from kh_common.utilities.json import json_stream
from kh_common.utilities import int_to_bytes
from kh_common.base64 import b64encode
from uuid import UUID, uuid4
from math import ceil
from time import time
import ujson as json


private_key = Ed25519PrivateKey.generate()
public_key = private_key.public_key().public_bytes(
	encoding=serialization.Encoding.DER,
	format=serialization.PublicFormat.SubjectPublicKeyInfo,
)
pk_signature = private_key.sign(public_key)

expires = int(time() + 1000)
issued = time()


def mock_pk(mocker, key_id=1) -> None :
	mocker.patch(
		'kh_common.auth.requests_post',
		side_effect=lambda *a, **kv : (
			None if kv['json'] != { 'key_id': key_id, 'algorithm': 'ed25519' } or a != (f'{auth_host}/v1/key',)
			else MockResponse({
				'signature': b64encode(pk_signature).decode(),
				'key': b64encode(public_key).decode(),
				'algorithm': 'ed25519',
				'expires': expires,
				'issued': issued,
			})
		)
	)


def mock_token(user_id: int, token_data:dict={ }, version:bytes=b'1', key_id=1, guid:UUID=None, valid_signature=True) -> str :
	load = b'.'.join([
		b'ed25519',                                        # algorithm
		b64encode(int_to_bytes(key_id)),                   # key_id
		b64encode(int_to_bytes(expires)),                  # expires timestamp
		b64encode(int_to_bytes(user_id)),                  # user id
		b64encode(guid.bytes if guid else uuid4().bytes),  # guid
		json.dumps(json_stream(token_data)).encode(),      # load
	])

	content = b64encode(version) + b'.' + b64encode(load)
	return (content + b'.' + b64encode(private_key.sign(content + (b'' if valid_signature else b'-')))).decode()



from datetime import datetime, timezone
from enum import Enum, unique
from typing import Dict
from os import environ

@unique
class Environment(Enum) :
	local: str = 'local'
	dev: str = 'dev'
	prod: str = 'prod'

	def is_local(self) :
		return self == Environment.local

	def is_dev(self) :
		return self == Environment.dev

	def is_prod(self) :
		return self == Environment.prod


environment: Environment = Environment[environ.get('ENVIRONMENT', 'LOCAL').lower()]

local: Dict[str, str] = {
	'auth_host': 'http://127.0.0.1:5000',
	'upload_host': 'http://localhost:5001',
	'tags_host': 'http://localhost:5002',
	'posts_host': 'http://localhost:5003',
	'account_host': 'http://localhost:5004',
	'users_host': 'http://localhost:5005',
	'config_host': 'http://localhost:5006',
}

dev: Dict[str, str] = {
	'auth_host': 'https://dev.kheina.com/auth',
	'upload_host': 'https://dev.kheina.com/upload',
	'tags_host': 'https://dev.kheina.com/tags',
	'posts_host': 'https://dev.kheina.com/posts',
	'account_host': 'https://dev.kheina.com/acct',
	'users_host': 'https://dev.kheina.com/users',
	'config_host': 'https://dev.kheina.com/config',
}

prod: Dict[str, str] = {
	'auth_host': 'https://auth.kheina.com',
	'upload_host': 'https://upload.kheina.com',
	'tags_host': 'https://tags.kheina.com',
	'posts_host': 'https://posts.kheina.com',
	'account_host': 'https://account.kheina.com',
	'users_host': 'https://users.kheina.com',
	'config_host': 'https://config.kheina.com',
}

assert local.keys() == dev.keys() == prod.keys()

env_vars: Dict[str, str] = locals().get(environment.name, local)

# add the variables from the environment to the module
locals().update(env_vars)

# delete extraneous data
del local, dev, prod, env_vars, environ


# put other variables/constants here (these will overwrite the env-specific configs above!)
epoch = 1576242000



import ujson


class MockResponse :

	def __init__(self, load) :
		if isinstance(load, str) :
			self.text = self.content = text
		elif isinstance(load, dict) :
			self.text = ujson.dumps(load)


	def json(self) :
		return ujson.loads(self.text)



from kh_common.utilities import stringSlice
from typing import Any, Dict, Union
from subprocess import PIPE, Popen


name: Union[str, None] = None

output: Union[bytes, None] = b''.join(Popen(['git', 'config', '--get', 'remote.origin.url'], stdout=PIPE, stderr=PIPE).communicate())
if output and not output.startswith(b'fatal'):
	name = stringSlice(output.decode(), '/', '.git')

else :
	output = b''.join(Popen(['git', 'rev-parse', '--show-toplevel'], stdout=PIPE, stderr=PIPE).communicate())
	if output and not output.startswith(b'fatal'):
		name = stringSlice(output.decode(), '/').strip()


short_hash: Union[str, None] = None

output = b''.join(Popen(['git', 'rev-parse', '--short', 'HEAD'], stdout=PIPE, stderr=PIPE).communicate())
if output and not output.startswith(b'fatal'):
	short_hash = output.decode().strip()


full_hash: Union[str, None] = None

output = b''.join(Popen(['git', 'rev-parse', 'HEAD'], stdout=PIPE, stderr=PIPE).communicate())
if output and not output.startswith(b'fatal'):
	full_hash = output.decode().strip()


del output, stringSlice, PIPE, Popen



from kh_common.sql import SqlInterface
from typing import Any, Hashable


class Map(SqlInterface, dict) :

	def __init__(self, table: str, key: str, value: str) :
		assert table.startswith('kheina.public.')
		SqlInterface.__init__(self)
		self._table = table
		self._key = key
		self._value = value


	def __missing__(self, key: Hashable) -> Any:
		data: Any = self.query(f"""
			SELECT {self._value}
			FROM {self._table}
			WHERE {self._key} = %s
			LIMIT 1;
			""",
			(key,),
			fetch_one=True,
		)
		self[key] = data[0]
		return data[0]



from signal import signal, SIGINT, SIGTERM


class Terminated:

	alive = True
	_run_on_term = []

	def on_terminate(func, *args, **kwargs) :
		Terminated._run_on_term.append((func, args, kwargs))

	def terminate(signum, frame):
		for func, args, kwargs in Terminated._run_on_term :
			func(*args, **kwargs)

		Terminated._run_on_term.clear()
		Terminated.alive = False

	def __call__(self = None) :
		return not Terminated.alive

	signal(SIGINT, terminate)
	signal(SIGTERM, terminate)



from aiohttp import ClientResponse, ClientTimeout, request as request_async
from requests import Response, get as requests_get, post as requests_post
from asyncio import get_event_loop, sleep as sleep_async
from kh_common.exceptions.base_error import BaseError
from kh_common.config.repo import name, short_hash
from kh_common.logging import getLogger, Logger
from kh_common.config.credentials import b2
from hashlib import sha1 as hashlib_sha1
from urllib.parse import quote, unquote
from typing import Any, Dict, Union
from base64 import b64encode
from time import sleep
import ujson as json


class B2AuthorizationError(BaseError) :
	pass


class B2UploadError(BaseError) :
	pass


class B2Interface :

	def __init__(self, timeout:float=300, max_backoff:float=30, max_retries:float=15, mime_types:Dict[str, str]={ }) -> None :
		self.logger: Logger = getLogger()
		self.b2_timeout: float = timeout
		self.b2_max_backoff: float = max_backoff
		self.b2_max_retries: float = max_retries
		self.mime_types: Dict[str, str] = {
			'jpg': 'image/jpeg',
			'jpeg': 'image/jpeg',
			'png': 'image/png',
			'webp': 'image/webp',
			'gif': 'image/gif',
			'webm': 'video/webm',
			'mp4': 'video/mp4',
			'mov': 'video/quicktime',
			**mime_types,
		}
		self._b2_authorize()


	def _b2_authorize(self) -> bool :
		basic_auth_string: bytes = b'Basic ' + b64encode((b2['key_id'] + ':' + b2['key']).encode())
		b2_headers: Dict[str, bytes] = { 'Authorization': basic_auth_string }
		response: Union[Response, None] = None

		for _ in range(self.b2_max_retries) :
			try :
				response = requests_get(
					'https://api.backblazeb2.com/b2api/v2/b2_authorize_account',
					headers=b2_headers,
					timeout=self.b2_timeout,
				)

			except :
				pass

			else :
				if response.ok :
					self.b2: Dict[str, Any] = json.loads(response.content)
					self.b2['upload_url_load']: Dict[str, str] = { 'bucketId': self.b2['allowed']['bucketId'] }
					return True

		else :
			raise B2AuthorizationError(
				'b2 authorization handshake failed.',
				response=json.loads(response.content) if response else None,
				status=response.status_code if response else None,
			)


	def _get_mime_from_filename(self, filename: str) -> str :
		extension: str = filename[filename.rfind('.') + 1:]
		if extension in self.mime_types :
			return self.mime_types[extension.lower()]
		raise ValueError(f'file extention does not have a known mime type: {filename}')


	def _obtain_upload_url(self) -> Dict[str, Any] :
		backoff: float = 1
		content: Union[str, None] = None
		status: Union[int, None] = None

		for _ in range(self.b2_max_retries) :
			try :
				response = requests_post(
					self.b2['apiUrl'] + '/b2api/v2/b2_get_upload_url',
					json=self.b2['upload_url_load'],
					headers={ 'Authorization': self.b2['authorizationToken'] },
					timeout=self.b2_timeout,
				)
				if response.ok :
					return json.loads(response.content)

				elif response.status_code == 401 :
					# obtain new auth token
					self._b2_authorize()

				else :
					content = response.content
					status = response.status_code

			except Exception as e :
				self.logger.warning('error encountered during b2 obtain upload url.', exc_info=e)

			sleep(backoff)
			backoff = min(backoff * 2, self.b2_max_backoff)

		raise B2AuthorizationError(
			f'Unable to obtain b2 upload url, max retries exceeded: {self.b2_max_retries}.',
			response=json.loads(content) if content else None,
			status=status,
		)


	async def _obtain_upload_url_async(self) -> Dict[str, Any] :
		backoff: float = 1
		content: Union[str, None] = None
		status: Union[int, None] = None

		for _ in range(self.b2_max_retries) :
			try :
				async with request_async(
					'POST',
					self.b2['apiUrl'] + '/b2api/v2/b2_get_upload_url',
					json=self.b2['upload_url_load'],
					headers={ 'Authorization': self.b2['authorizationToken'] },
					timeout=ClientTimeout(self.b2_timeout),
				) as response :
					if response.ok :
						return await response.json()

					elif response.status == 401 :
						# obtain new auth token
						self._b2_authorize()

					else :
						content = await response.read()
						status = response.status

			except Exception as e :
				self.logger.warning('error encountered during b2 obtain upload url.', exc_info=e)

			await sleep_async(backoff)
			backoff = min(backoff * 2, self.b2_max_backoff)

		raise B2AuthorizationError(
			f'Unable to obtain b2 upload url, max retries exceeded: {self.b2_max_retries}.',
			response=json.loads(content) if content else None,
			status=status,
		)


	def b2_upload(self, file_data: bytes, filename: str, content_type:Union[str, None]=None, sha1:Union[str, None]=None) -> Dict[str, Any] :
		# obtain upload url
		upload_url: str = self._obtain_upload_url()

		sha1: str = sha1 or hashlib_sha1(file_data).hexdigest()
		content_type: str = content_type or self._get_mime_from_filename(filename)

		headers: Dict[str, str] = {
			'Authorization': upload_url['authorizationToken'],
			'X-Bz-File-Name': quote(filename),
			'Content-Type': content_type,
			'Content-Length': str(len(file_data)),
			'X-Bz-Content-Sha1': sha1,
		}

		backoff: float = 1
		content: Union[str, None] = None
		status: Union[int, None] = None

		for _ in range(self.b2_max_retries) :
			try :
				response = requests_post(
					upload_url['uploadUrl'],
					headers=headers,
					data=file_data,
					timeout=self.b2_timeout,
				)
				status = response.status_code
				if response.ok :
					content: Dict[str, Any] = json.loads(response.content)
					assert content_type == content['contentType']
					assert sha1 == content['contentSha1']
					assert filename == unquote(content['fileName'])
					return content

				else :
					content = response.content

			except AssertionError :
				raise

			except Exception as e :
				self.logger.warning('error encountered during b2 upload.', exc_info=e)

			sleep(backoff)
			backoff = min(backoff * 2, self.b2_max_backoff)

		raise B2UploadError(
			f'Upload to b2 failed, max retries exceeded: {self.b2_max_retries}.',
			response=json.loads(content) if content else None,
			status=status,
			upload_url=upload_url,
			headers=headers,
			filesize=len(file_data),
		)


	async def b2_upload_async(self, file_data: bytes, filename: str, content_type:Union[str, None]=None, sha1:Union[str, None]=None) -> Dict[str, Any] :
		# obtain upload url
		upload_url: str = await self._obtain_upload_url_async()

		sha1: str = sha1 or hashlib_sha1(file_data).hexdigest()
		content_type: str = content_type or self._get_mime_from_filename(filename)

		headers: Dict[str, str] = {
			'Authorization': upload_url['authorizationToken'],
			'X-Bz-File-Name': quote(filename),
			'Content-Type': content_type,
			'Content-Length': str(len(file_data)),
			'X-Bz-Content-Sha1': sha1,
		}

		backoff: float = 1
		content: Union[str, None] = None
		status: Union[int, None] = None

		for _ in range(self.b2_max_retries) :
			try :
				async with request_async(
					'POST',
					upload_url['uploadUrl'],
					headers=headers,
					data=file_data,
					timeout=ClientTimeout(self.b2_timeout),
				) as response :
					status = response.status
					if response.ok :
						content: Dict[str, Any] = await response.json()
						assert content_type == content['contentType']
						assert sha1 == content['contentSha1']
						assert filename == unquote(content['fileName'])
						return content

					else :
						content = await response.read()

			except AssertionError :
				raise

			except Exception as e :
				self.logger.warning('error encountered during b2 upload.', exc_info=e)

			await sleep_async(backoff)
			backoff = min(backoff * 2, self.b2_max_backoff)

		raise B2UploadError(
			f'Upload to b2 failed, max retries exceeded: {self.b2_max_retries}.',
			response=json.loads(content) if content else None,
			status=status,
			upload_url=upload_url,
			headers=headers,
			filesize=len(file_data),
		)



from kh_common.utilities import getFullyQualifiedClassName
from kh_common.exceptions.base_error import BaseError
from kh_common.logging import getLogger, Logger
from fastapi.responses import UJSONResponse
from typing import Dict, Union
from fastapi import Request
from uuid import uuid4


logger: Logger = getLogger()


def jsonErrorHandler(request: Request, e: Exception) -> UJSONResponse :
	status: int = getattr(e, 'status', 500)

	error: Dict[str, Union[str, int]] = {
		'status': status,
		'refid': getattr(e, 'refid', uuid4()).hex,
	}

	if isinstance(e, BaseError) :
		error['error'] = f'{e.__class__.__name__}: {e}'

	else :
		logger.error(error, exc_info=e)
		error['error'] = 'Internal Server Error'

	return UJSONResponse(
		error,
		status_code=status,
	)



from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey
from cryptography.hazmat.primitives.serialization import load_der_public_key
from kh_common.exceptions.http_error import Forbidden, Unauthorized
from typing import Any, Callable, Dict, NamedTuple, Set, Union
from cryptography.hazmat.backends import default_backend
from kh_common.models import AuthToken, KhUser, Scope
from kh_common.config.constants import auth_host
from kh_common.utilities import int_from_bytes
from requests import post as requests_post
from kh_common.caching import ArgsCache
from datetime import datetime, timezone
from kh_common.base64 import b64decode
from enum import Enum, unique
from fastapi import Request
from uuid import UUID
import ujson as json


class KhUser(KhUser) :
	def authenticated(self, raise_error=True) :
		if not self.token or self.token != verifyToken(self.token.token_string) :
			if raise_error :
				raise Unauthorized('User is not authenticated.')
			return False
		return True

	def verify_scope(self, scope: Scope) :
		self.authenticated()
		if scope not in self.scope :
			raise Forbidden('User is not authorized to access this resource.')
		return True


@ArgsCache(60 * 60 * 24)  # 24 hour cache
def _fetchPublicKey(key_id: int, algorithm: str) -> Dict[str, Union[str, int, Ed25519PublicKey]] :
	response: Response = requests_post(f'{auth_host}/v1/key', json={ 'key_id': key_id, 'algorithm': algorithm })
	load: Dict[str, Union[str, float, int]] = json.loads(response.text)

	signature: bytes = b64decode(load['signature'])
	key: bytes = b64decode(load['key'])
	public_key: Ed25519PublicKey = load_der_public_key(key, backend=default_backend())

	# don't verify in try/catch so that it doesn't cache an invalid token
	public_key.verify(signature, key)

	return {
		'key_id': key_id,
		'algorithm': algorithm,
		'public_key': public_key,
		'expires': datetime.fromtimestamp(load['expires'], timezone.utc),
	}


def v1token(token: str) -> AuthToken :
	content: str
	signature: str
	version: str
	load: str

	content, signature = token.rsplit('.', 1)
	version, load = tuple(map(b64decode, content.split('.')))

	algorithm: bytes
	key_id: bytes
	expires: bytes
	user_id: bytes
	guid: bytes
	data: bytes

	algorithm, key_id, expires, user_id, guid, data = load.split(b'.', 5)

	algorithm: str = algorithm.decode()
	key_id: int = int_from_bytes(b64decode(key_id))
	expires: datetime = datetime.fromtimestamp(int_from_bytes(b64decode(expires)), timezone.utc)
	user_id: int = int_from_bytes(b64decode(user_id))
	guid: UUID = UUID(bytes=b64decode(guid))

	if datetime.now(timezone.utc) > expires :
		raise Unauthorized('Key has expired.')

	key: Dict[str, Union[str, int, Ed25519PublicKey]] = _fetchPublicKey(key_id, algorithm)

	if datetime.now(timezone.utc) > key['expires'] :
		raise Unauthorized('Key has expired.')

	try :
		key['public_key'].verify(b64decode(signature), content.encode())

	except :
		raise Unauthorized('Key validation failed.')

	return AuthToken(
		guid=guid,
		user_id=user_id,
		expires=expires,
		data=json.loads(data),
		token_string=token,
	)


tokenVersionSwitch: Dict[bytes, Callable] = {
	b'1': v1token,
}


def verifyToken(token: str) -> AuthToken :
	version: bytes = b64decode(token[:token.find('.')])

	if version in tokenVersionSwitch :
		return tokenVersionSwitch[version](token)

	raise ValueError('The given token uses a version that is unable to be decoded.')


def retrieveAuthToken(request: Request) -> AuthToken :
	token: str = request.headers.get('Authorization') or request.cookies.get('kh-auth')

	if not token :
		raise Unauthorized('An authentication token was not provided.')

	token_data: AuthToken = verifyToken(token.split()[-1])

	# TODO: this works kind of weird with ipv6 and ephemeral ip addresses, fix later
	# if 'ip' in token_data.data and token_data.data['ip'] != request.client.host :
	# 	raise Unauthorized('The authentication token provided is not valid from this device or location.')

	return token_data



from typing import Any, Dict, List, Tuple, Union
from uuid import UUID, uuid4


class BaseError(Exception) :
	def __init__(self, message: str, *args:Tuple[Any], refid:Union[UUID, str, None]=None, logdata:Dict[str, Any]={ }, **kwargs: Dict[str, Any]) -> None :
		Exception.__init__(self, message)

		self.refid: str = refid or logdata.get('refid') or uuid4()

		if isinstance(self.refid, (str, bytes)) :

			if len(self.refid) == 32 :
				self.refid = UUID(hex=self.refid)

			elif len(self.refid) == 16 :
				self.refid = UUID(bytes=self.refid)

			else :
				raise ValueError('badly formed refid.')

		if 'refid' in logdata :
			del logdata['refid']

		self.logdata: Dict[str, Any] = {
			**logdata,
			**kwargs,
		}



from typing import Any, Dict, NamedTuple, Set
from datetime import datetime
from enum import Enum, unique
from uuid import UUID


class AuthToken(NamedTuple) :
	user_id: int
	expires: datetime
	guid: UUID
	data: Dict[str, Any]
	token_string: str


@unique
class Scope(Enum) :
	default: int = 0
	bot: int = 1
	user: int = 2
	mod: int = 3
	admin: int = 4

	def all_included_scopes(self) :
		return [v for v in Scope.__members__.values() if Scope.user.value <= v.value <= self.value] or [self]


class KhUser(NamedTuple) :
	user_id: int
	token: AuthToken
	scope: Set[Scope]

	def __hash__(self) :
		return hash(f'{self.user_id}{self.scope}')



from re import compile as re_compile, _pattern_type as Pattern
from typing import Callable, Dict, Iterable, Iterator, List
from collections import defaultdict


_tagOperators: Dict[str, Callable] = defaultdict(lambda : normalizeTag, {
	'/a': str.lower,
	'/r': str.lower,
	'/l': lambda x : x[:-2],
})

_repetitionCount = 3

_repetitionRegex: Pattern = re_compile(r'(\w+?)\1{' + str(_repetitionCount) +  r',}')

_tagReplaceRegex: Pattern = re_compile(r'[-_]+')


def normalizeTag(tag: str) -> str :
	return _repetitionRegex.sub(lambda x: x.group(1) * _repetitionCount, _tagReplaceRegex.sub('_', '_'.join(tag.split())).strip('-_').lower())


def stripTags(tags: Iterable[str]) -> Iterator[str] :
	for tag in tags :
		yield tag
		tag = tag.rstrip('s')
		yield tag
		yield tag.rstrip('e')


def tagSplit(tags: List[str]) -> Iterator[str] :
	return map(lambda x : _tagOperators[x[-2:]](x), filter(None, map(str.strip, tags.split(','))))



from kh_common.config.constants import epoch
from kh_common.caching import CalcDict
from scipy.stats import norm
from math import log10, sqrt
from typing import Union


"""
resources:
	https://github.com/reddit-archive/reddit/blob/master/r2/r2/lib/db/_sorts.pyx
	https://steamdb.info/blog/steamdb-rating
	https://www.evanmiller.org/how-not-to-sort-by-average-rating.html
	https://redditblog.com/2009/10/15/reddits-new-comment-sorting-system
	https://www.reddit.com/r/TheoryOfReddit/comments/bpmd3x/how_does_hot_vs_best_vscontroversial_vs_rising/envijlj
"""


z_score = CalcDict(lambda k : norm.ppf(1-(1-k)/2))


def _sign(x: Union[int, float]) -> int :
	return (x > 0) - (x < 0)


def hot(up: int, down: int, time: float) -> float :
	s: int = up - down
	return _sign(s) * log10(max(abs(s), 1)) + (time - epoch) / 45000


def controversial(up: int, down: int) -> float :
	return (up + down)**(min(up, down)/max(up, down)) if up or down else 0


def confidence(up: int, total: int, z:float=z_score[0.8]) -> float :
	if not total :
		return 0
	phat = up / total
	return (
		(phat + z * z / (2 * total)
		- z * sqrt((phat * (1 - phat)
		+ z * z / (4 * total)) / total)) / (1 + z * z / total)
	)


def best(up: int, total: int) -> float :
	if not total :
		return 0
	s: float = up / total
	return s - (s - 0.5) * 2**(-log10(total + 1))



from asyncio import ensure_future, Queue, QueueEmpty, sleep, wait as WaitAll
from aiohttp import ClientTimeout, request as async_request
from kh_common.config.credentials import telegram
from kh_common.utilities.signal import Terminated
from kh_common.caching import Aggregate
from kh_common.logging import getLogger
from collections import defaultdict
from inspect import getfullargspec
from typing import List


logger = getLogger()


class QuitParsing(Exception) :
	pass


class Listener :

	def __init__(self,
		loop_time: float = 1,
		queue_empty_wait: float = 1,
		threads: int = 1,
		allow_chats: bool = False,
		bot_name: str = None,
		timeout: float = 30,
		text_limit: int = 4096,
		text_breaks: List[str] = ['\n', '\t', ' '],
		# commands that don't need to run any logic
		responses: dict = { },
		# commands that actually require logic to be performed
		commands: dict = { },
	) :
		self.loop_time = loop_time
		self.queue_empty_wait = queue_empty_wait
		self.allow_chats = allow_chats
		self.timeout = timeout
		self.threads = threads
		self.bot_name = bot_name.lower() if bot_name else None

		self._telegram_access_token = telegram['telegram_access_token']
		self._telegram_bot_id = telegram['telegram_bot_id']

		self.commands = defaultdict(lambda : self.handleUnknownCommand, commands)
		self.responses = responses
		self.text_limit = text_limit
		self.text_breaks = text_breaks

		self.queue = Queue()


	def splitMessage(self, text: str) :
		if len(text) <= self.text_limit :
			return [text]

		messages = []

		while len(text) > self.text_limit :
			t = text[:self.text_limit]
			t_split = -1

			for splitter in self.text_breaks :
				t_split = t.rfind(splitter)

				if t_split > 0 :
					break

			if t_split <= 0 :
				t_split = self.text_limit
				splitter = ''

			messages.append(text[:t_split])
			text = text[t_split + len(splitter):]

		messages.append(text)
		return messages


	# parse_mode = MarkdownV2 or HTML
	async def sendMessage(self, recipient, message, parse_mode='HTML') :
		messages = self.splitMessage(message)
		success = False

		for i, m in enumerate(messages) :
			success = await self._sendSingleMessage(recipient, m, parse_mode, i, len(messages))
			if not success :
				return False

		return success


	async def _sendSingleMessage(self, recipient, message, parse_mode='HTML', message_index=0, message_count=1) :
		request = f'https://api.telegram.org/bot{self._telegram_access_token}/sendMessage'
		error = 'failed to send notification to telegram.'
		info = None
		for _ in range(5) :
			try :
				async with async_request(
					'POST',
					request,
					json={
						'parse_mode': parse_mode,
						'chat_id': recipient,
						'text': message,
					},
					timeout=ClientTimeout(self.timeout),
				) as response :
					info = await response.json()
					if not info['ok'] :
						break
					return True
			except :
				pass

		logger.error({
			'info': info,
			'message': error,
			'request': {
				'url': request,
				'message': {
					'text': message,
					'index': message_index,
					'total': message_count,
				},
			},
		})
		return False


	async def handleNonCommand(self, chat, is_chat, **kwargs) :
		if not is_chat :
			ensure_future(self._sendSingleMessage(chat, 'Sorry, I only understand bot commands right now.'))


	async def handleUnknownCommand(self, chat, **kwargs) :
		ensure_future(self._sendSingleMessage(chat, "Sorry, I didn't understand that command. to see a list of my commands, try /help"))


	async def parseMessage(self, message) :
		user = message['from']['id']
		chat = message['chat']['id']
		is_chat = False

		if user != chat :
			if not self.allow_chats :
				return True

			is_chat = True

		args = {
			'user': user,
			'chat': chat,
			'is_chat': is_chat,
			'message': message,
		}

		try :
 			entity = next(filter(lambda x : x['type'] == 'bot_command', message.get('entities', [])))

		except StopIteration :
			arg_spec = getfullargspec(self.handleNonCommand)

			if arg_spec.varkw :
				return await func(**args)

			return await self.handleNonCommand(**{ k: args[k] for k in arg_spec.args[1:] + arg_spec.kwonlyargs })

		end = entity['offset'] + entity['length']
		command = message['text'][entity['offset']:end]

		if is_chat :
			command_split = command.split('@')
			if len(command_split) <= 1 :
				return True

			if command_split[1].lower() != self.bot_name :
				return True

			command = command_split[0]

		if command in self.responses :
			return await self.sendMessage(chat, self.responses[command])

		args.update({
			'text': message['text'][end:].strip(),
			'command': command,
		})

		func = self.commands[command]
		arg_spec = getfullargspec(func)

		if arg_spec.varkw :
			return await func(**args)

		return await func(**{ k: args[k] for k in arg_spec.args[1:] + arg_spec.kwonlyargs })


	async def processQueue(self) :
		while Terminated.alive or not self.queue.empty :
			try :
				update = self.queue.get_nowait()

			except QueueEmpty :
				await sleep(self.queue_empty_wait)
				continue

			try :
				await self.parseMessage(update['message'])

			except QuitParsing :
				pass

			except :
				logger.exception({
					'message': 'failed to parse message.',
					'update': update,
				})

			self.queue.task_done()


	async def run(self) :
		threads = [self.processQueue() for _ in range(self.threads)] + [self.recv()]
		await WaitAll(threads)


	async def recv(self) :
		# just let it fail if it's not json serialized
		request = f'https://api.telegram.org/bot{self._telegram_access_token}/getUpdates?offset='
		mostrecent = 0
		while Terminated.alive :
			try :
				async with async_request(
					'GET',
					request + str(mostrecent),
					timeout=ClientTimeout(self.timeout),
				) as response :
					updates = await response.json()

					if not updates['ok'] :
						logger.error({
							'message': 'failed to read updates from telegram.',
							'updates': updates,
						})
						await sleep(self.loop_time)

					elif updates['result'] :
						mostrecent = updates['result'][-1]['update_id'] + 1
						for update in updates['result'] :
							await self.queue.put(update)

					else :
						await sleep(self.loop_time)

				self._logQueueSize(self.queue.qsize())

			except :
				logger.exception('failed to read updates from telegram.')


	@Aggregate(60)
	def _logQueueSize(self, queue_size: int) :
		logger.info({
			'queue_size': queue_size,
		})



from kh_common.models.privacy import Privacy
from pydantic import BaseModel
from datetime import datetime
from typing import Optional


class User(BaseModel) :
	name: str
	handle: str
	privacy: Privacy
	icon: str
	banner: Optional[str]
	website: Optional[str]
	created: datetime
	description: Optional[str]
	verified: Optional[str]



from inspect import FullArgSpec, getfullargspec, iscoroutinefunction
from typing import Any, Callable, Dict, Iterable, Set, Tuple
from kh_common.exceptions.base_error import BaseError
from kh_common.logging import getLogger, Logger
from uuid import UUID, uuid4
from functools import wraps
from enum import Enum


logger: Logger = getLogger()


class HttpError(BaseError) :
	def __init__(self, message: str, *args:Tuple[Any], status:int=500, **kwargs: Dict[str, Any]) -> None :
		super().__init__(message, *args, **kwargs)
		self.status = status


class NotFound(HttpError) :
	def __init__(self, message: str, *args:Tuple[Any], **kwargs: Dict[str, Any]) -> None :
		super().__init__(message, *args, status=404, **kwargs)


class BadRequest(HttpError) :
	def __init__(self, message: str, *args:Tuple[Any], **kwargs: Dict[str, Any]) -> None :
		super().__init__(message, *args, status=400, **kwargs)


class Unauthorized(HttpError) :
	def __init__(self, message: str, *args:Tuple[Any], **kwargs: Dict[str, Any]) -> None :
		super().__init__(message, *args, status=401, **kwargs)


class Forbidden(HttpError) :
	def __init__(self, message: str, *args:Tuple[Any], **kwargs: Dict[str, Any]) -> None :
		super().__init__(message, *args, status=403, **kwargs)


class UnsupportedMedia(HttpError) :
	def __init__(self, message: str, *args:Tuple[Any], **kwargs: Dict[str, Any]) -> None :
		super().__init__(message, *args, status=415, **kwargs)


class UnprocessableEntity(HttpError) :
	def __init__(self, message: str, *args:Tuple[Any], **kwargs: Dict[str, Any]) -> None :
		super().__init__(message, *args, status=422, **kwargs)


class Conflict(HttpError) :
	def __init__(self, message: str, *args:Tuple[Any], **kwargs: Dict[str, Any]) -> None :
		super().__init__(message, *args, status=409, **kwargs)


class InternalServerError(HttpError) :
	pass


class ResponseNotOk(HttpError) :
	pass


class BadOrMalformedResponse(HttpError) :
	pass


def HttpErrorHandler(message: str, exclusions: Iterable[str] = ['self'], handlers: Dict[type, Tuple[type, str]] = { }) -> Callable :
	"""
	raises internal server error from any unexpected errors
	f'an unexpected error occurred while {message}.'
	"""

	exclusions: Set[str] = set(exclusions)

	def decorator(func: Callable) -> Callable :

		arg_spec: FullArgSpec = getfullargspec(func)

		if iscoroutinefunction(func) :
			@wraps(func)
			async def wrapper(*args: Tuple[Any], **kwargs:Dict[str, Any]) -> Any :
				try :
					return await func(*args, **kwargs)

				except HttpError :
					raise

				except Exception as e :
					for cls in type(e).__mro__ :
						if cls in handlers :
							Error, custom_message = handlers[cls]
							raise Error(custom_message)

					kwargs.update(zip(arg_spec.args, args))
					kwargs['refid']: UUID = uuid4().hex

					logdata = {
						key: kwargs[key]
						for key in kwargs.keys() - exclusions
					}
					logger.exception({ 'params': logdata })

					raise InternalServerError(f'an unexpected error occurred while {message}.', logdata=logdata)

		else :
			@wraps(func)
			def wrapper(*args: Tuple[Any], **kwargs:Dict[str, Any]) -> Any :
				try :
					return func(*args, **kwargs)

				except HttpError :
					raise

				except Exception as e :
					for cls in type(e).__mro__ :
						if cls in handlers :
							Error, custom_message = handlers[cls]
							raise Error(custom_message)

					kwargs.update(zip(arg_spec.args, args))
					kwargs['refid']: UUID = uuid4().hex

					logdata = {
						key: kwargs[key]
						for key in kwargs.keys() - exclusions
					}
					logger.exception({ 'params': logdata })

					raise InternalServerError(f'an unexpected error occurred while {message}.', logdata=logdata)

		return wrapper

	return decorator



from kh_common import __version__ as kh_version
from setuptools import setup, find_packages


setup(
	name='kh_common',
	version=kh_version,
	description='common libraries for kheina.com',
    long_description=open('readme.md').read(),
    long_description_content_type='text/markdown',
	author='kheina',
	url='https://github.com/kheina-com/kh-common',
	packages=find_packages(exclude=['tests']),
	install_requires=open('requirements.txt').read().split(),
	extras_require={
		'scoring':  ['scipy>=1.5.2'],
	},
)


from starlette.types import ASGIApp, Receive, Send, Message, Scope as request_scope
from kh_common.auth import AuthToken, KhUser, retrieveAuthToken, Scope
from kh_common.exceptions.http_error import HttpError, BadRequest
from starlette.datastructures import Headers, MutableHeaders
from kh_common.exceptions import jsonErrorHandler
from starlette.requests import Request
from fastapi.responses import Response
from urllib.parse import urlparse
from functools import partial
from typing import Iterable

class KhCorsMiddleware:

	def __init__(
		self,
		app: ASGIApp,
		allowed_origins: Iterable[str],
		allowed_protocols: Iterable[str] = ['https'],
		allowed_headers: Iterable[str] = [],
		allowed_methods: Iterable[str] = [],
		allow_credentials: bool = True,
		exposed_headers: Iterable[str] = [],
		max_age: int=86400,
	) -> None :
		self.app = app
		self.allowed_origins = set(allowed_origins)
		self.allowed_protocols = set(allowed_protocols)
		self.allowed_headers = ', '.join(allowed_headers + ['access-control-request-method', 'origin'])
		self.allowed_methods = ', '.join(map(str.upper, allowed_methods))
		self.allow_credentials = str(allow_credentials).lower()
		self.exposed_headers = ', '.join(exposed_headers)
		self.max_age = str(max_age)


	async def __call__(self, scope: request_scope, receive: Receive, send: Send) -> None :
		if scope['type'] != 'http' :
			await self.app(scope, receive, send)
			return

		request: Request = Request(scope, receive, send)

		if 'origin' in request.headers :
			origin = urlparse(request.headers['origin'])

			if origin.scheme not in self.allowed_protocols or origin.netloc.split(':')[0] not in self.allowed_origins :
				response = jsonErrorHandler(request, BadRequest('Origin not allowed.'))
				await response(scope, receive, send)
				return

			if request.method == 'OPTIONS' and 'access-control-request-method' in request.headers :
				await Response(
					None,
					status_code=204,
					headers={
						'access-control-allow-origin': origin.geturl(),
						'access-control-allow-methods': self.allowed_methods,
						'access-control-allow-headers': self.allowed_headers,
						'access-control-allow-credentials': self.allow_credentials,
						'access-control-max-age': self.max_age,
						'access-control-expose-headers': self.exposed_headers,
					},
				)(scope, receive, send)
				return

			send = partial(self.send, send=send, headers=request.headers)

		await self.app(scope, receive, send)


	async def send(self, message: Message, send: Send, headers: Headers) -> None :
		if message['type'] != 'http.response.start':
			await send(message)
			return

		origin = headers['origin']
		message.setdefault('headers', [])
		headers = MutableHeaders(scope=message)

		headers.update({
			'access-control-allow-origin': origin,
			'access-control-allow-methods': self.allowed_methods,
			'access-control-allow-headers': self.allowed_headers,
			'access-control-allow-credentials': self.allow_credentials,
			'access-control-max-age': self.max_age,
			'access-control-expose-headers': self.exposed_headers,
		})

		await send(message)



from aiohttp import BasicAuth, ClientTimeout, request as async_request
from kh_common.exceptions.base_error import BaseError
from kh_common.config.credentials import mailgun
from kh_common.logging import getLogger
from dataclasses import dataclass
from asyncio import sleep
from uuid import uuid4


_html_template_1 = "<!DOCTYPE html><html lang='en'><head><style>body{height:100%;width:100%;position:absolute;background:#C3C4CE;background-size:cover;background-position:center;}body,html{background:#C3C4CE;position:relative;z-index:-5;margin:0;padding:0;font-family:Bitstream Vera Sans,DejaVu Sans,Arial,Helvetica,sans-serif;}a,form input,form label,.footer span{cursor:pointer;pointer-events:all;text-decoration:none;color:#222222;transition: ease 0.15s;}a:link{color:#222222;}a:visited{color:inherit;}a:hover{color:#F28817!important;opacity:1!important;transition: ease 0.15s;}h1{margin:0 0 25px;}p{margin:0;}#content{display:block;margin:100px auto;width:100%;padding:25px 0;text-align:center;background:#E0E4E8;}#feature{display:block;margin:0 auto;max-width:900px;padding:0;background:#E0E4E8;}.button{display:inline-block;padding:0.5em 1em;margin:25px 25px 0;border:1px solid #2D333A;background:#D8D9E0; box-shadow:0 2px 3px 1px #6D718680;border-radius:3px;white-space:nowrap;}.button:hover{box-shadow:0 0 10px 3px #6D7186B3;border-color:#F28817;}.subtext{color:#00000080;margin:25px 0 0;font-size:0.7em;}</style></head>"
_html_template_2 = "<!-- ♀ --><body><div id='content'><main id='feature'>{title}<p>{text}</p>{button}{subtext}</main></div></body></html>"
logger = getLogger()


class EmailError(BaseError) :
	pass


@dataclass
class Button :
	link: str
	text: str


def formatHtml(text:str, title:str=None, button: Button=None, subtext:str=None) :
	return _html_template_1 + _html_template_2.format(
		text=text,
		title=f'<h1>{title}</h1>' if title else '',
		button=f"<a class='button' href='{button.link}'>{button.text}</a>" if button else '',
		subtext=f"<p class='subtext'>{subtext}</p>" if subtext else '',
	)


def formatText(text:str, title:str=None, button: Button=None, subtext:str=None) :
	if title :
		text = title + '\n\n' + text
	if button :
		text += f'\n\n{button.text}: {button.link}'
	if subtext :
		text += '\n\n' + subtext
	return text


async def sendEmail(
	to: str,
	subject:str,
	text:str,
	title:str=None,
	button: Button=None,
	subtext: str=None,
	sender:str='kheina.com <system@kheina.com>',
	cc:str=None,
	bcc:str=None,
	timeout:int=30,
) :
	html = formatHtml(text, title, button, subtext)
	text = formatText(text, title, button, subtext)

	payload = {
		'from': sender,
		'to': to,
		'subject': subject,
		'text': text,
		'html': html,
	}

	if cc :
		payload['cc'] = cc

	if bcc :
		payload['bcc'] = bcc

	for i in range(5) :
		try :
			async with async_request(
				'POST',
				mailgun['endpoint'],
				auth=BasicAuth(**mailgun['auth']),
				data=payload,
				timeout=ClientTimeout(timeout),
				raise_for_status=True,
			) as response :
				return True
		except :
			await sleep(i)

	guid = uuid4()
	logdata = {
		'message': 'failed to send email.',
		'email': {
			'to': to,
			'subject': subject,
		},
		'refid': guid.hex,
	}
	logger.critical(logdata)
	raise EmailError('failed to send email.', refid=guid, logdata=logdata)



from kh_common.caching import ArgsCache
from kh_common.sql import SqlInterface
from kh_common.hashing import Hashable
from typing import Set


class UserBlocking(SqlInterface, Hashable) :

	def __init__(self) :
		Hashable.__init__(self)
		SqlInterface.__init__(self)


	@ArgsCache(60)
	def user_blocked_tags(self, user_id: int) -> Set[str] :
		data = self.query(
			"""
			SELECT tags.tag
			FROM kheina.public.tag_blocking
				INNER JOIN kheina.public.tags
					ON tags.tag_id = blocked
						AND tags.deprecated = false
			WHERE user_id = %s;
			""",
			(user_id,),
			fetch_all=True,
		)

		return set(data)


	@ArgsCache(60)
	def user_blocked_users(self, user_id: int) -> Set[str] :
		data = self.query(
			"""
			SELECT users.handle
			FROM kheina.public.user_blocking
				INNER JOIN kheina.public.users
					ON users.user_id = blocked
			WHERE user_id = %s;
			""",
			(user_id,),
			fetch_all=True,
		)

		return set(data)
